#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <WebSocketsServer.h>
#include <ESP8266mDNS.h>
#include <ArduinoJson.h>
#include "FS.h"
#include <OneWire.h>
#include <DallasTemperature.h>

#define NAME "DS18B20" 

const char *ssid = "ESP8266-" NAME;
const char *password = "";

//DS18B20
#define ONE_WIRE_BUS 5
#define ONE_WIRE_MAX_DEV 1

OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature DS18B20(&oneWire);
int numberOfDevices;
DeviceAddress devAddr[ONE_WIRE_MAX_DEV];
float tempDev[ONE_WIRE_MAX_DEV];
float tempDevLast[ONE_WIRE_MAX_DEV];
long lastTemp;
const int durationTemp = 5000;

//Convert device id to String
String GetAddressToString(DeviceAddress deviceAddress)
{
  String str = "";
  for (uint8_t i = 0; i < 8; i++){
    if (deviceAddress[i] < 16) {
      str += String(0, HEX);
    }
    str += String(deviceAddress[i], HEX);
  }
  return str;
}

//Setting the temperature sensor
void SetupDS18B20()
{
  DS18B20.begin();

  Serial.print("Parasite power is: "); 
  if (DS18B20.isParasitePowerMode()) { 
    Serial.println("ON");
  } else {
    Serial.println("OFF");
  }
  
  numberOfDevices = DS18B20.getDeviceCount();
  Serial.print("Device count: ");
  Serial.println(numberOfDevices);

  lastTemp = millis();
  DS18B20.requestTemperatures();

  // Loop through each device, print out address
  for (int i = 0; i < numberOfDevices; i++){
    // Search the wire for address
    if (DS18B20.getAddress(devAddr[i], i)) {
      //devAddr[i] = tempDeviceAddress;
      Serial.print("Found device ");
      Serial.print(i, DEC);
      Serial.print(" with address: " + GetAddressToString(devAddr[i]));
      Serial.println();
    } else {
      Serial.print("Found ghost device at ");
      Serial.print(i, DEC);
      Serial.print(" but could not detect address. Check power and cabling");
    }

    //Get resolution of DS18b20
    Serial.print("Resolution: ");
    Serial.print(DS18B20.getResolution(devAddr[i]));
    Serial.println();

    //Read temperature from DS18b20
    float tempC = DS18B20.getTempC(devAddr[i]);
    Serial.print("Temp C: ");
    Serial.println(tempC);
  }
}

ESP8266WebServer server(80);
WebSocketsServer ws = WebSocketsServer(8080);

void ws_handle(uint8_t client, WStype_t type, uint8_t *payload, size_t length)
{
    static const int bufferSize = 128;
    static char buffer[bufferSize];
  
    switch(type) {
        case WStype_CONNECTED:
            break;
        case WStype_TEXT: {
            float tempC = DS18B20.getTempC(devAddr[0]);
            DS18B20.setWaitForConversion(false);
            DS18B20.requestTemperatures();
            snprintf(buffer, bufferSize, "%f", tempC);
            ws.sendTXT(client, buffer);
            break;
        }
    }
}

const char* wifi_ssid;
const char* wifi_psk;

bool loadConfig()
{
  File configFile = SPIFFS.open("/config.json", "r");
  if (!configFile) {
    Serial.println("Failed to open config file");
    return false;
  }

  size_t size = configFile.size();
  if (size > 1024) {
    Serial.println("Config file size is too large");
    return false;
  }

  std::unique_ptr<char[]> buf(new char[size]);
  configFile.readBytes(buf.get(), size);

  StaticJsonDocument<1024> doc;
  DeserializationError error = deserializeJson(doc, buf.get());
  if (error) {
    Serial.print("Failed to parse config file. Error: ");
    Serial.println(error.c_str());
    return false;
  }

  wifi_ssid = doc["ssid"];
  wifi_psk = doc["psk"];

  Serial.print("wifi_ssid: ");
  Serial.println(wifi_ssid);
  Serial.print("wifi_psk: ");
  Serial.println(wifi_psk);
  
  return true;
}

void ap_start()
{
    Serial.println("Configuring access point...");
    WiFi.softAP(ssid, password);
  
    IPAddress ip = WiFi.softAPIP();
    Serial.print("AP IP address: ");
    Serial.println(ip);
}

void setup()
{
  Serial.begin(115200);
  Serial.println();
  Serial.println(NAME " " __DATE__ " " __TIME__);
  delay(1000); // warm up
  
  Serial.println("Mounting FS...");
  
  if (!SPIFFS.begin()) {
    Serial.println("Failed to mount file system");
    return;
  }

  if (!loadConfig()) {
    Serial.println("Failed to load config");
    ap_start();
  } else {
    Serial.println("Config loaded");
    
    Serial.print("Connecting to ");
    Serial.println(ssid);

    WiFi.mode(WIFI_STA);
    WiFi.begin(wifi_ssid, wifi_psk);

    int i = 10;
    while (WiFi.status() != WL_CONNECTED) {
      delay(1000);
      i--;
      if (i <= 0) break;
      Serial.print(".");
    }

    if (WiFi.status() == WL_CONNECTED) {
      Serial.println("WiFi connected");
      Serial.println("IP address: ");
      Serial.println(WiFi.localIP());
    } else {
      Serial.println("WiFi connection failed");
      ap_start();
    }
  }

  //Setup DS18b20 temperature sensor
  SetupDS18B20();

  // start webSocket server
  ws.begin();
  ws.onEvent(ws_handle);
  
  server.on("/", HTTP_GET, []() {
    server.send(200, "text/html",
"<!DOCTYPE html><html><head>"
"<script>var connection = new WebSocket('ws://'+location.hostname+':8080/', [ 'arduino' ]);</script>"
"</head><body>"
"</body></html>");
  });
  
  server.on("/info", HTTP_GET, []() {
    server.send(200, "application/json", "{ \"name\": \"" NAME "\", \"websocket_port\": 8080 }");
  });
  
  server.on("/config", HTTP_GET, []() {
    server.send(200, "application/json", "{}");
  });
  server.on("/config/edit", HTTP_GET, []() {
    server.send(200, "text/html",
"<!DOCTYPE html><html><head>"
"</head><body>"
"</body></html>");
  });
  server.on("/config", HTTP_POST, []() {
    if (server.hasArg("ssid")) {
      wifi_ssid = server.arg("ssid").c_str();
    }
    if (server.hasArg("psk")) {
      wifi_psk = server.arg("psk").c_str();
    }
    
    StaticJsonDocument<1024> doc;
    doc["ssid"] = wifi_ssid;
    doc["psk"] = wifi_psk;
  
    File configFile = SPIFFS.open("/config.json", "w");
    if (configFile) {
      serializeJson(doc, configFile);
      server.send(200);
    } else {
      Serial.println("Failed to open config file for writing");
      server.send(500);
    }
  });

  server.on("/", HTTP_POST, []() {
  });
  
  server.begin();
  Serial.println("HTTP server started");
}

void loop()
{
  ws.loop();
  server.handleClient();
}
